Red/System [
	Title:		"ZeroMQ Binding"
	Author:		"Kaj de Vos"
	Rights:		"Copyright (c) 2011 Kaj de Vos. All rights reserved."
	License: {
		Redistribution and use in source and binary forms, with or without modification,
		are permitted provided that the following conditions are met:

		    * Redistributions of source code must retain the above copyright notice,
		      this list of conditions and the following disclaimer.
		    * Redistributions in binary form must reproduce the above copyright notice,
		      this list of conditions and the following disclaimer in the documentation
		      and/or other materials provided with the distribution.

		THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
		ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
		WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
		DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
		FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
		DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
		SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
		CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
		OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
		OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	}
	Needs:		"0MQ >= 2.0.7"
	Notes:		"Switch the zmq_init call in the new-pool routine for 0MQ <= 2.0.6"
	Tabs:		4
]


#define none		null
#define none?		[null = ]

#define yes			true
#define no			false

#define handle!		[pointer! [integer!]]


; System interface

#define error-again	11  ; EAGAIN: Linux, Windows; TODO: check for other platforms


; 0MQ interface

#import [
	"libzmq.dll" cdecl [
		version: "zmq_version" [  ; Return 0MQ version.
			major		[struct! [value [integer!]]]
			minor		[struct! [value [integer!]]]
			patch		[struct! [value [integer!]]]
		]

		new-pool: "zmq_init" [  ; Return context handle.
;			app-threads	[integer!]  ; For 0MQ <= 2.0.6
			io-threads	[integer!]
;			flags		[integer!]  ; For 0MQ <= 2.0.6
			return: 	[pointer!]
		]
		end-pool: "zmq_term" [  ; Clean up context.
			pool		[handle!]
			return: 	[integer!]
		]

		open: "zmq_socket" [  ; Open a socket.
			pool		[handle!]
			type		[integer!]
			return: 	[pointer!]
		]
		close: "zmq_close" [  ; Clean up a socket from a context.
			socket		[handle!]
			return: 	[integer!]
		]

		set: "zmq_setsockopt" [  ; "Set socket option."
			socket		[handle!]
			name		[integer!]
			value		[pointer!]
			size		[integer!]
			return: 	[integer!]
		]
		; For 0MQ > 2.0.6:
		option: "zmq_getsockopt" [  ; "Get socket option."
			socket		[handle!]
			name		[integer!]
			value		[pointer!]  ; Currently max 255 bytes
			size		[struct! [value [integer!]]]
			return: 	[integer!]
		]

		serve: "zmq_bind" [  ; Set up server socket binding.
			socket		[handle!]
			end-point	[c-string!]
			return: 	[integer!]
		]
		connect: "zmq_connect" [  ; Connect to a server socket.
			socket		[handle!]
			destination	[c-string!]
			return: 	[integer!]
		]

		new-message: "zmq_msg_init_size" [  ; Create new message.
			message		[struct! message!]
			size		[integer!]
			return: 	[integer!]
		]
		clear-message: "zmq_msg_init" [  ; Initialise new message.
			message		[struct! message!]
			return: 	[integer!]
		]
		as-message: "zmq_msg_init_data" [  ; Convert to new message.
			message		[struct! message!]
			data		[c-string!]
			size		[integer!]
			free		[handle!]
			hint		[handle!]
			return: 	[integer!]
		]
		end-message: "zmq_msg_close" [  ; Clean up message.
			message		[struct! message!]
			return: 	[integer!]
		]

		message-data: "zmq_msg_data" [  ; Return message data pointer.
			message		[struct! message!]
			return:		[c-string!]
		]
		message-size: "zmq_msg_size" [  ; Return message data size.
			message		[struct! message!]
			return: 	[integer!]
		]

		send-message: "zmq_send" [  ; Send message.
			socket		[handle!]
			message		[struct! message!]
			flags		[integer!]
			return: 	[integer!]
		]
		receive-message: "zmq_recv" [  ; Receive a message.
			socket		[handle!]
			message		[struct! message!]
			flags		[integer!]
			return: 	[integer!]
		]

		copy-message: "zmq_msg_copy" [  ; "Copy message content to another message."
			target		[pointer!]
			source		[pointer!]
			return: 	[integer!]
		]
		move-message: "zmq_msg_move" [  ; "Move message content to another message."
			target		[pointer!]
			source		[pointer!]
			return: 	[integer!]
		]

		wait: "zmq_poll" [  ; "Wait for selected events or timeout."
			events		[handle!]
			length		[integer!]
			timeout		[integer!]
			return: 	[integer!]
		]

		error: "zmq_errno" [  ; Return last status.
			return: 	[integer!]
		]
		error-message: "zmq_strerror" [  ; Return status message.
			code		[integer!]
			return:		[c-string!]
		]
	]
]

; open types

#define pair				0
#define publish				1
#define subscribe			2
#define request				3
#define reply				4
#define extended-request	5
#define extended-reply		6
#define push				7
#define pull				8

; Socket options

#define max-messages		1
;#define min-messages		2
#define swap-size			3
#define io-affinity		4
#define identity			5
#define filter				6
#define unsubscribe		7
#define max-rate			8
#define recovery-interval	9
#define loop-back?			10
#define send-buffer		11
#define receive-buffer		12

; send/receive flags

#define no-block			1

; wait designation masks

#define poll-in				1
#define poll-out			2
#define poll-error			4


; Higher level interface

; FIXME: not thread safe
message!: alias struct! [
	content		[handle!]
	flags		[byte!]  ; unsigned char
	vsm-size	[byte!]  ; unsigned char
	vsm-data1	[integer!]  ; unsigned char [ZMQ_MAX_VSM_SIZE] (default 30)
	vsm-data2	[integer!]
	vsm-data3	[integer!]
	vsm-data4	[integer!]
	vsm-data5	[integer!]
	vsm-data6	[integer!]
	vsm-data7	[integer!]
	vsm-data8	[integer!]
]
message: struct message!

send: func [  ; Send message.
	socket		[handle!]
	data		[c-string!]
	size		[integer!]
	flags		[integer!]
	return: 	[logic!]
;	/local		message [handle!]
][
	either zero? as-message message data size none none [
		either zero? send-message socket message flags [
			zero? end-message message
		][
			end-message message
			no
		]
	][
		no
	]
]

receive: func [  ; Receive a message.
	socket		[handle!]
;	data		[handle!]  ; Double reference
;	size		[pointer! [integer!]]
	flags		[integer!]
;	return: 	[logic!]
	return: 	[handle!]
;	/local		message [handle!]
][
	either zero? clear-message message [
		either zero? receive-message socket message flags [
;			size: message-size message
;			copy-memory data  message-data message  size
;			data/string: message-data message
;			data: message-data message

;			zero? end-message message
			message  ; FIXME: seems to be invalid
		][
			end-message message
			none
		]
	][
		none
	]
]
